# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
#
# Invenio App RDM is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Files utilities."""

from __future__ import absolute_import, print_function

from datetime import datetime

import sqlalchemy as sa
from flask import current_app
from flask_mail import Message
from invenio_files_rest.models import FileInstance
from invenio_rdm_records.records.models import RDMDraftMetadata, RDMRecordMetadata
from pytz import utc


def checksum_verification_files_query():
    """Return a FileInstance query taking into account file URI prefixes."""
    files = FileInstance.query
    uri_prefixes = current_app.config.get(
        "FILES_REST_CHECKSUM_VERIFICATION_URI_PREFIXES"
    )
    if uri_prefixes:
        files = files.filter(
            sa.or_(*[FileInstance.uri.startswith(p) for p in uri_prefixes])
        )
    return files


def render_email_from_context(template, context):
    """Renders a template using a provided context."""
    template = current_app.jinja_env.get_or_select_template(template)
    return template.render(context)


def generate_integrity_report_entries(files):
    """Generates the report entries."""
    entries = []
    for file in files:
        entry = _generate_report_entry(file)
        entries.append(entry)
    return entries


def _generate_report_entry(file):
    """Generates a single report for a file."""
    entry = {"file": file}
    for o in file.objects:
        entry["filename"] = o.key

        # Find records/drafts for the files
        draft = get_draft_from_bucket(o.bucket_id)
        if draft and draft.json:
            entry["draft"] = draft.json
        else:
            record = get_record_from_bucket(o.bucket_id)
            if record and record.json:
                entry["record"] = record.json
    return entry


def get_record_from_bucket(bucket_id):
    """Retrieve a record from a bucket id."""
    return RDMRecordMetadata.query.filter_by(bucket_id=bucket_id).one_or_none()


def get_draft_from_bucket(bucket_id):
    """Retrieve a draft from a bucket id."""
    return RDMDraftMetadata.query.filter_by(bucket_id=bucket_id).one_or_none()


def get_report_subject():
    """Retrieves the integrity report e-mail's subject."""
    subject = current_app.config.get(
        "APP_RDM_FILES_INTEGRITY_REPORT_SUBJECT", "Files integrity report"
    )
    now = datetime.now(tz=utc).strftime("%Y-%m-%d %H:%M:%S")
    subject += f" [{now}]"
    return subject


def get_report_template():
    """Retrieves the template name to be used in the files integrity report."""
    tpl = current_app.config["APP_RDM_FILES_INTEGRITY_REPORT_TEMPLATE"]
    return tpl


def render_report_email(files):
    """Renders the report e-mail from a list of files."""
    entries = generate_integrity_report_entries(files)
    tpl = get_report_template()
    context = {"entries": entries}
    return render_email_from_context(tpl, context)


def send_integrity_report_email(files):
    """Sends an e-mail with a report on the provided files.

    APP configs ('MAIL_DEFAULT_SENDER', 'APP_RDM_ADMIN_EMAIL_RECIPIENT') must be set.
    Oherwise a warning is logged.
    """
    try:
        if files:
            subject = get_report_subject()
            body = render_report_email(files)
            sender = current_app.config["MAIL_DEFAULT_SENDER"]
            admin_email = current_app.config["APP_RDM_ADMIN_EMAIL_RECIPIENT"]
            recipients = admin_email
            if not type(admin_email) == list:
                recipients = [admin_email]
            mail_ext = current_app.extensions["mail"]
            msg = Message(subject, sender=sender, recipients=recipients, body=body)
            mail_ext.send(msg)
    except Exception as e:
        current_app.logger.error("Integrity report not sent. Error: {}".format(str(e)))



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 CERN.
# Copyright (C) 2021 Northwestern University.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Test files-rest is protected."""

from io import BytesIO


def create_draft(client, record, headers):
    """Create draft and return its id."""
    response = client.post("/records", json=record, headers=headers)
    assert response.status_code == 201
    return response.json["id"]


def init_file(client, recid, headers):
    """Init a file for draft with given recid."""
    return client.post(
        f"/records/{recid}/draft/files", headers=headers, json=[{"key": "test.pdf"}]
    )


def upload_file(client, recid):
    """Create draft and return its id."""
    return client.put(
        f"/records/{recid}/draft/files/test.pdf/content",
        headers={
            "content-type": "application/octet-stream",
            "accept": "application/json",
        },
        data=BytesIO(b"testfile"),
    )


def commit_file(client, recid, headers):
    """Create draft and return its id."""
    return client.post(f"/records/{recid}/draft/files/test.pdf/commit", headers=headers)


# NOTE: It seems like it was already the case that a logged in user wouldn't be
#       able to access files-rest. We are just making doubly-clear.
def test_files_rest_endpoint_is_protected(
    running_app, client_with_login, headers, es_clear, minimal_record
):
    client = client_with_login

    # Create draft with file
    minimal_record["files"] = {"enabled": True}
    recid = create_draft(client, minimal_record, headers)
    init_file(client, recid, headers)
    upload_file(client, recid)
    commit_file(client, recid, headers)

    # Get bucket information
    url = f"/records/{recid}/draft/files/test.pdf"
    response = client.get(url, headers=headers)
    bucket_id = response.json["bucket_id"]

    # Nobody is allowed to use the invenio-files-rest endpoints
    # (even logged-in user). Just testing for the GET of each is enough

    bucket_url = f"/files/{bucket_id}"
    response = client.get(bucket_url, headers=headers)
    assert 404 == response.status_code  # because of files-rest hiding feature

    bucket_key_url = f"/files/{bucket_id}/test.pdf"
    response = client.get(bucket_key_url, headers=headers)
    assert 404 == response.status_code



# -*- coding: utf-8 -*-
#
# Copyright (C) 2019-2021 CERN.
# Copyright (C) 2019-2021 Northwestern University.
#
# Invenio-RDM-Records is free software; you can redistribute it and/or modify
# it under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

import pytest
from invenio_pidstore.models import PersistentIdentifier

SINGLE_RECORD_API_URL = "/records/{}"
LIST_RECORDS_API_URL = "/records"
DRAFT_API_URL = "/records/{}/draft"
DRAFT_ACTION_API_URL = "/records/{}/draft/actions/{}"


def test_record_read_non_existing_pid(client, location, minimal_record, es_clear):
    """Retrieve a non existing record."""
    # retrieve unknown record
    response = client.get(SINGLE_RECORD_API_URL.format("notfound"))
    assert response.status_code == 404
    assert response.json["status"] == 404
    assert response.json["message"] == "The persistent identifier does not exist."


def test_record_draft_create_and_read(
    client_with_login, running_app, minimal_record, es_clear
):
    """Test draft creation of a non-existing record."""
    # create a record
    client = client_with_login
    response = client.post(LIST_RECORDS_API_URL, json=minimal_record)

    assert response.status_code == 201

    response_fields = response.json.keys()
    fields_to_check = ["id", "metadata", "revision_id", "created", "updated", "links"]

    for field in fields_to_check:
        assert field in response_fields

    recid = response.json["id"]

    # retrieve record draft
    response = client.get(DRAFT_API_URL.format(recid))
    assert response.status_code == 200
    assert response.json is not None


def test_record_draft_publish(
    client_with_login, headers, running_app, minimal_record, es_clear
):
    """Test draft publication of a non-existing record.

    It has to first create said draft and includes record read.
    """
    # Create the draft
    client = client_with_login
    response = client.post(LIST_RECORDS_API_URL, json=minimal_record, headers=headers)

    assert response.status_code == 201
    recid = response.json["id"]

    # Publish it
    response = client.post(
        DRAFT_ACTION_API_URL.format(recid, "publish"), headers=headers
    )

    assert response.status_code == 202
    response_fields = response.json.keys()
    fields_to_check = ["id", "metadata", "revision_id", "created", "updated", "links"]

    for field in fields_to_check:
        assert field in response_fields

    response = client.get(DRAFT_API_URL.format(recid), headers=headers)
    assert response.status_code == 404

    # Test record exists
    response = client.get(SINGLE_RECORD_API_URL.format(recid), headers=headers)

    assert response.status_code == 200

    response_fields = response.json.keys()
    fields_to_check = ["id", "metadata", "revision_id", "created", "updated", "links"]

    for field in fields_to_check:
        assert field in response_fields


def test_read_record_with_redirected_pid(
    client_with_login, headers, running_app, minimal_record, es_clear
):
    """Test read a record with a redirected pid."""
    # Create dummy record
    client = client_with_login
    response = client.post(LIST_RECORDS_API_URL, headers=headers, json=minimal_record)
    assert response.status_code == 201
    # Publish it
    pid1_value = response.json["id"]
    response = client.post(
        DRAFT_ACTION_API_URL.format(pid1_value, "publish"), headers=headers
    )
    assert response.status_code == 202

    # Create another dummy record
    response = client.post(LIST_RECORDS_API_URL, headers=headers, json=minimal_record)
    assert response.status_code == 201
    pid2_value = response.json["id"]
    # Publish it
    response = client.post(
        DRAFT_ACTION_API_URL.format(pid2_value, "publish"), headers=headers
    )
    assert response.status_code == 202

    # redirect pid1 to pid2
    pid1 = PersistentIdentifier.get("recid", pid1_value)
    pid2 = PersistentIdentifier.get("recid", pid2_value)
    pid1.redirect(pid2)

    response = client.get(SINGLE_RECORD_API_URL.format(pid1.pid_value), headers=headers)
    assert response.status_code == 301

    assert response.json["status"] == 301
    assert response.json["message"] == "Moved Permanently."


@pytest.mark.skip()
def test_read_deleted_record(
    client_with_login, headers, location, minimal_record, es_clear, admin_user
):
    """Test read a deleted record."""
    client = client_with_login

    # Create dummy record to test delete
    response = client.post(LIST_RECORDS_API_URL, headers=headers, json=minimal_record)
    assert response.status_code == 201
    recid = response.json["id"]
    # Publish it
    response = client.post(
        DRAFT_ACTION_API_URL.format(recid, "publish"), headers=headers
    )
    assert response.status_code == 202

    # Delete the record
    response = client.delete(SINGLE_RECORD_API_URL.format(recid), headers=headers)
    assert response.status_code == 204

    # Read the deleted record
    response = client.get(SINGLE_RECORD_API_URL.format(recid), headers=headers)
    assert response.status_code == 410
    assert response.json["message"] == "The record has been deleted."


def test_record_search(client, headers, running_app, es_clear):
    """Test record search."""
    expected_response_keys = set(["hits", "links", "aggregations"])
    expected_metadata_keys = set(["resource_type", "creators", "titles"])

    # Get published bibliographic records
    response = client.get(LIST_RECORDS_API_URL, headers=headers)

    assert response.status_code == 200
    response_keys = set(response.json.keys())
    # The datamodel has other tests (jsonschemas, mappings, schemas)
    # Here we just want to crosscheck the important ones are there.
    assert expected_response_keys.issubset(response_keys)

    for r in response.json["hits"]["hits"]:
        metadata_keys = set(r["metadata"])
        assert expected_metadata_keys.issubset(metadata_keys)
